import numpy as np
from numpy import pi, sqrt, sin, cos
#from matplotlib import pyplot as plt

def perifocal(coe, f):
    """
    Given the COE for an orbit and the angle (f) from perisapsis (in
    radians), then this returns the (p,q) cartesian locationl

    https://www.sciencedirect.com/topics/engineering/perifocal-frame
    """
    p = coe.a * (1 - coe.e ** 2)
    pos = np.array([cos(f), sin(f)]) * p / (1 + coe.e * cos(f))
    return pos

def orbit(coe):
    """
    Produces the (x,y) or (p,q) locations with in the perifocal plan (always 2D)
    of the orbit. Allows a simple 2D plot in jupyter notebook with:

        r = orbit(coe)
        plt.plot(r[:,0], r[:,1],'--r')
        plt.grid(True);

    coe: Classical Orbital Elements
    return: numpy([(x,y),(x,y), ...]) which has a shape of (360,2)

    https://www.sciencedirect.com/topics/engineering/perifocal-frame
    """
    orb = np.zeros((360,2))
    deg2rad = pi/180
    for i,f in enumerate(range(360)):
        orb[i,:] = perifocal(coe, f*deg2rad)

    return orb




import numpy as np
from numpy import cos, sqrt, sin, pi
import attr

from .tle import parse_tle
from .constants import Earth, deg2rad
from rotations.rotations import R1,R2,R3,R313

degree_symbol = u"\xb0"

@attr.s()
class COE:
    a = attr.ib()
    e = attr.ib()
    i = attr.ib()
    raan = attr.ib()
    w = attr.ib()
    v = attr.ib()

    tle = attr.ib(default=None)

    def to_rv(self):
        return coe2rv(self.a, self.e, self.i, self.raan, self.w, self.v)

    @staticmethod
    def from_rv(R, V):
        """Returns the orbit in R(x,y,z)[km] and V(x,y,z)[km/sec]"""
        a, e, i, node, w, v = rv2coe(R, V)
        return COE(a, e, i, node, w, v)

    @staticmethod
    def from_tle(tle):
        a = parse_tle(tle)
        c = COE(*a.coe)
        c.tle = a
        return c

    @property
    def r(self):
        """Returns the scalar radius[km]"""
        return self.a*(1-self.e**2)/(1+self.e*cos(self.v*pi/180))

    @property
    def velocity(self):
        """Returns the scalar velocity[km/sec]"""
        mu = Earth.mu
        # non-circular orbit
        if self.e > 1e-6:
            return sqrt(mu*(2/self.r-1/self.a))
        # circular orbit
        return sqrt(mu/self.r)

    @property
    def period(self):
        mu = Earth.mu
        return 2*pi*sqrt(self.a**3/mu)

    def __str__(self):
        s = f"a: {self.a:.1f}km e: {self.e:.4f} i: {self.i:.1f}\xb0 RAAN: {self.raan:.1f}\xb0 w: {self.w:.1f}\xb0 v: {self.v:.1f}\xb0"
        return s


######################################################################

def rv2coe(R, V, mu=Earth.mu):
    """Given postion (R) and velocity (V) in an ECI frame, this returns
    the classical orbital elements: (a, e, i, node, w, v)
    """
    H = np.cross(R,V)

    v = np.linalg.norm(V)
    r = np.linalg.norm(R)
    energy = v*v/2 - mu/r
    a = -(mu/2)/(v*v/2-mu/r)

    E = np.cross(V,H)/mu - R/r
    e = np.linalg.norm(E)

    i = np.arccos(H[2]/np.linalg.norm(H))
    if i < 1e-5: # equitorial orbit
        node = 0
        if e < 1e-5:  # circular
            vv = R[0]/r if R[1] > 0 else 2*pi-R[0]/r
            w = np.NaN
        else: # elliptical
            w = np.arccos(E[0]/e) if E[1] > 0 else 2*pi - np.arccos(E[0]/e)
            vv = np.arccos(np.dot(E,R)/(e*r))
            if np.dot(R,V) < 0:
                vv = 2*pi - vv
    else:
        N = np.cross(np.array([0,0,1]), H)
        n = np.linalg.norm(N)

        node = np.arccos(N[0]/n)
        node = 2*pi-node if N[1] <0 else node

        w = np.arccos(np.dot(N,E)/(d*e))
        w = 2*pi-w if E[2]<0 else w

    return (a, e, i*180/pi, node*180/pi, w*180/pi, vv*180/pi)


def coe2rv(a, e, i, node, w, v, MU=Earth.mu, degrees=True):
    """Given the classical orbital elements (a, e, i, node, w, v), this
    returns the position (R) and the velocity (V) in an ECI frame

    - Semimajor-axis (a)[km]: orbit size
    - Eccentricity (e): orbit shape (0=circle, 1=line)
    - Inclination (i)[deg]: orbital plane inclination measure from ascending node
    - Argument of Perigee (w)[deg]: orbit orientation
    - Ascending Node (Omega)[deg]: location of ascending node
    - True Anomaly (v)[deg]: location of satellite in orbit relative to perigee
    - Mean Anomaly (M)[deg]: fictitious angle that varies linearly with time

    return: R(x,y,z)[km], V(x,y,z)[km/sec]
    """
    # MU = MU/1000/1000/1000  # FIXME?

    if degrees:
        i *= deg2rad
        node *= deg2rad
        w *= deg2rad
        v *= deg2rad

    p = a*(1-e**2)  # p = semi-latus rectum (semiparameter)
    R = np.zeros(3)
    V = np.zeros(3)
    sv = sin(v)
    cv = cos(v)
    det = 1/(1+e*cv)
    smup = np.sqrt(MU/p)

    ###  Position Coordinates in Perifocal Coordinate System
    # R[0] = p*cv / (1+e*cv) # x-coordinate (km)
    # R[1] = p*sv / (1+e*cv) # y-coordinate (km)
    # R[2] = 0                             # z-coordinate (km)
    # V[0] = -sqrt(MU/p) * sv       # velocity in x (km/s)
    # V[1] =  sqrt(MU/p) * (e+cv)   # velocity in y (km/s)
    # V[2] =  0                            # velocity in z (km/s)


    R[0] = p*cv * det     # x-coordinate (km)
    R[1] = p*sv * det     # y-coordinate (km)
    R[2] = 0              # z-coordinate (km)
    V[0] = -smup * sv     # velocity in x (km/s)
    V[1] =  smup * (e+cv) # velocity in y (km/s)
    V[2] =  0             # velocity in z (km/s)

    r313 = R313(-node, -i, -w)

    # Perifocal -> xyz
    R = r313.dot(R)
    V = r313.dot(V)

    return (R,V,)



from math import pi, pow
from datetime import datetime, timedelta, timezone
from doop.constants import Earth
from collections import namedtuple

from enum import IntFlag

ID = namedtuple("ID", "launch_year launch_number piece")
Object = namedtuple("Object", "name number classification")

EphemerisType = IntFlag("EphemerisType","SGP SGP4 SDP4 SGP8 SDP8")

TLE = namedtuple("TLE",
        'object '
        'id '
        'coe '
        'ballistic_coeffecient '
        'bstar '
        'line1 line2'
    )

OE = namedtuple("OE", "a e i raan w v")

def fix_classification(x):
    if x == "U":
        return "Unclassified"
    elif x == "C":
        return "Classified"
    elif x == "S":
        return "Secret"
    raise Exception(f"Invalid classification: {x}")

def fix_sci(x):
    """
    TLE have a shorthand for storing scientific numbers, just fixing it
    """
    # leading +/- can trip this up
    if x[0] == '+' or x[0] == '-':
        xx = x[1:]
    else:
        xx = x

    s = xx.split("-")
    ss = xx.split("+")
    if len(s) == 2:
        x = float(s[0] + "e-" + s[1])
    elif len(ss) == 2:
        s = ss
        x = float(s[0] + "e+" + s[1])
    else:
        x = float(xx)
    return x

def fix_dec(x):
    """
    TLE formate assumes a leading decimal point, just putting it back in
    """
    if x[0] == "-":
        ret = float("-0." + x[1:])
    else:
        ret = float("0." + x)
    return ret

def fix_year(year:int):
    # FIXME: need a better solution :)
    return 2000+year if year < 60 else 1900+year

def fix_epoch(day:float, year:int):
    """https://ubuntuforums.org/archive/index.php/t-2032246.html"""
    # print(f"day {day} year {year}")
    yr = fix_year(year)
    return datetime(year=yr, month=1, day=1, tzinfo=timezone.utc) + timedelta(days=day-1)


def parse_tle(tle:str):
    """
    Format
    https://en.wikipedia.org/wiki/Two-line_element_set
    https://celestrak.com/NORAD/documentation/tle-fmt.php
    https://spaceflight.nasa.gov/realdata/sightings/SSapplications/Post/JavaSSOP/SSOP_Help/tle_def.html

    AAAAAAAAAAAAAAAAAAAAAAAA
    1 NNNNNU NNNNNAAA NNNNN.NNNNNNNN +.NNNNNNNN +NNNNN-N +NNNNN-N N NNNNN
    2 NNNNN NNN.NNNN NNN.NNNN NNNNNNN NNN.NNNN NNN.NNNN NN.NNNNNNNNNNNNNN

    ISS (ZARYA)
    1 25544U 98067A   20060.61908565  .00000737  00000-0  21434-4 0  9993
    2 25544  51.6436 165.6500 0005418 332.6966 228.1099 15.49204316215186

    TLE are in ECI frame
    """
    # name, onel, twol = tle.split("\n")
    tle.strip()
    if tle[-1] == "\n":
        tle = tle[:-1]

    tmp = tle.split("\n")
    if len(tmp) == 2:
        name = "unknown"
        onel = tmp[0]
        twol = tmp[1]
    elif len(tmp) >= 3:
        name, onel, twol = tmp[:3]
    else:
        print(tmp)
        raise Exception(f"tle_parse: too many lines: {len(tmp)}")

    one = onel.split()
    two = twol.split()
    # print(f"name: {name}\n one: {one}\n two: {two}\n")

    cat = str(one[1]).strip()
    cl = fix_classification(cat[-1])
    obj = Object(name.strip(), int(cat[:-1]),cl)

    yr = fix_year(int(one[2][:2]))
    piece = str(one[2][-1])
    num = str(one[2][2:-1])
    id = ID(yr, num, piece)

    yr = int(one[3][:2])
    day = float(one[3][2:])
    # print("epoch", fix_epoch(day, yr))

    n = float(two[7][:11])  # mean motion, or period of orbit
    u = Earth.mu

    a = pow(u,1/3)/pow(2*n*pi/86400, 2/3)
    e = fix_dec(two[4])
    i = float(two[2])
    raan = float(two[3])
    w = float(two[5])
    v = float(two[6])

    # classic orbital elements: a e i raan w v
    coe = OE(a,e,i,raan,w,v)

    d = int(one[7])
    if d > 0:
        print(EphemerisType(d))
    # else:
    #     print(f"unknown ephemeris type: {d}")

    return TLE(
            obj,
            id,
            coe,
            # float(two[63:68]), # rev num at epoch - this number isn't useful/accurate
            float(one[4]), # ballistic coeff
            fix_sci(one[6]), # bstar
            # n, # mean motion
            # fix_sci(one[44:52]), # 2nd der mean motion
            onel,
            twol
        )




import requests
from .elements import COE
from colorama import Fore

# from collections import namedtuple

# SatelliteTLE = {
#     "Dragon": """DRAGON CRS-2
#     1 39115U 13010A   13062.62492353  .00008823  00000-0  14845-3 0   188
#     2 39115  51.6441 272.5899 0012056 334.2535  68.5574 15.52501943   306""",
#
#     "ISS": """ISS (ZARYA)
#     1 25544U 98067A   08264.51782528 -.00002182  00000-0 -11606-4 0  2927
#     2 25544  51.6416 247.4627 0006703 130.5360 325.0288 15.72125391563537""",
#
#     "DragonDemo": """CREW DRAGON DEMO-1
#     1 44063U 19011A   19062.34862786 -.10981604  00000-0 -22652+0 0  9998
#     2 44063  51.6335 167.4655 0006446 102.4656 353.9277 15.51500210   163""",
#
#     "ChinaSat": """CHINASAT 2D
#     1 43920U 19001A   19062.86574468 -.00000352  00000-0  00000+0 0  9995
#     2 43920   0.0402 293.1489 0001187  74.3893 235.4568  1.00273137   662"""
# }


class Celestrak:
    endpts = {
        "gps": "https://celestrak.com/NORAD/elements/gps-ops.txt",
        "galileo": "https://celestrak.com/NORAD/elements/galileo.txt",
        "beidou": "https://celestrak.com/NORAD/elements/beidou.txt",
        "glonass": "https://celestrak.com/NORAD/elements/glo-ops.txt",
        "stations": "https://celestrak.com/NORAD/elements/stations.txt",
        "starlink": "https://celestrak.com/NORAD/elements/starlink.txt"
    }

    def raw(self, w):
        if w not in self.endpts:
            print(f"{Fore.RED}{w} is invalid; choose: {list(w.keys())}{Fore.RESET}")
            raise Exception

        f = self.endpts[w]
        resp = requests.get(f)
        if resp.status_code != 200:
            raise Exception(f"{Fore.RED}Failed to get TLEs from: {f}{Fore.RESET}")

        # print(len(resp.text))
        # print(resp.text)

        d = resp.text.split("\r\n")
        # print(len(d))
        # print(d[:6])
        tles = []
        for i in range(0, len(d), 3):
            if d[i] is None:
                break
            s = "\n".join(d[i:i+3])
            tles.append(s)

        print(f">> Found {len(tles)} satellites")
        return tles

    def tles(self, w):
        pass

    def coes(self, w):
        tles = self.raw(w)

        coes = []
        for tle in tles:
            try:
                coes.append(COE.from_tle(tle))
            except Exception as e:
                print(f"{Fore.CYAN}{tle}{Fore.RESET}")
                print(f"{Fore.RED}*** {e} ***{Fore.RESET}")
        return coes



from collections import namedtuple
from math import pi

# value?
femto = 1E-15
pico = femto*1000
nano = pico*1000
micro = nano*1000
milli = micro*1000
centi = 1E-2
kilo = 1E3
mega = kilo*1000
giga = mega*1000

# conversions
deg2rad = pi/180
rad2deg = 180/pi

LagrangePoint = namedtuple("LagrangePoint","L1 L2 L3 L4 L5")
EarthMoonLagrange = LagrangePoint(326000e3,448900e3,381680e3,384400e3,384400e3) # m
SunEarthLagrange = LagrangePoint(148.11e9,151.1e9,149.6e9,149.6e9,149.6e9)
GEO = 42164e3 # m

speedOfLight = 3*10**8 # m/s

# FIXME: change grav const to km^3/sec^2?? orbits use km not m
# https://en.wikipedia.org/wiki/Standard_gravitational_parameter
# mu = GM
# G = standard gravitational parameter of celectial body
# M = mass of body

# radius [km]
# mass [kg]
# standard gravitational const [m^3/sec^2]
# tilt [deg]


Body = namedtuple("Body", "radius mass mu tilt geo")

to_km3 = 1/1000/1000/1000
# https://en.wikipedia.org/wiki/Earth
Earth = Body(6378.388, 5.97237E24, 3.986004418E14*to_km3, 23.4392811, 42164)

# https://en.wikipedia.org/wiki/Moon
Moon = Body(1737.4, 7.342e22, 4.9048695e12*to_km3, 1.5424, None)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['doop']

package_data = \
{'': ['*']}

install_requires = \
['numpy', 'rotations']

setup_kwargs = {
    'name': 'doop',
    'version': '0.0.5',
    'description': 'orbital dynamics',
    'long_description': '![](https://raw.githubusercontent.com/nimbus-bp-1729/doop/master/pics/zapp.png)\n\n![Python](https://github.com/nimbus-bp-1729/doop/workflows/Python/badge.svg)\n\n# DOOP\n\n**Still figuring out what I want to do with this**\n\nSome simple python code to do astrodynamics work in python.\n\n- hohman transfers\n- convert between classical orbital parameters and position/velocity in ECI\n- convert between some useful reference frames: \n    - Earth Central Inertial (ECI) [x,y,z]\n    - Earth Centered, Earth Fixed (ECEF) [x,y,z]\n    - North East Down (NED) [x,y,z]\n    - Geocentric [lat,lon,alt]\n\n## Documents\n\n### Two Line Elements ([TLEs](docs/tle/tle.md))\n\n```\nTBD\n```\n\n### Classical Orbital Elements\n\n![](https://upload.wikimedia.org/wikipedia/commons/thumb/e/eb/Orbit1.svg/266px-Orbit1.svg.png)\n\n- Semimajor-axis (a)[km]: orbit size\n- Eccentricity (e)[unitless]: orbit shape (0=circule, 1=straight line)\n- Inclination (i)[deg]: orbital plane inclination measure from ascending node\n- Argument of Perigee (w)[deg]: orbit orientation\n- Ascending Node (Omega)[deg]: location of ascending node\n- True Anomaly (v)[deg]: location of satellite in orbit relative to perigee\n- Mean Anomaly (M)[deg]: fictitious angle that varies linearly with time\n\n## Other Useful Packages\n\n- [spaceman3D](https://github.com/Jaseibert/spaceman3D) reading tles, plotting, appears dead, don\'t think it works correctly ... everything has same RAAN\n- [orbital](https://github.com/RazerM/orbital) reading tles, plotting, appears dead\n- [spg4](https://github.com/brandon-rhodes/python-sgp4) propagation, seems ok\n- [tle-tools](https://pypi.org/project/TLE-tools/) not sure value, seems dead\n- [ESA: pykep](https://esa.github.io/pykep/index.html) more for inter-planetary stuff\n- [astrodynamics](https://github.com/dinkelk/astrodynamics) looks good, but it is Matlab code\n- [OrPytal](https://github.com/nicklafarge/OrPytal) does orbits without specific parameters\n\n# References\n\n- [Wiki: Earth parameters](https://en.wikipedia.org/wiki/Earth)\n- [Wiki: Orbital elements](https://en.wikipedia.org/wiki/Orbital_elements)\n- [Wiki: TLE Format](https://en.wikipedia.org/wiki/Two-line_element_set)\n- [celestrak.com TLEs](https://celestrak.com/NORAD/elements/)\n\n# MIT License\n\n**Copyright (c) 2020 Kevin J. Walchko**\n\nPermission is hereby granted, free of charge, to any person obtaining a copy\nof this software and associated documentation files (the "Software"), to deal\nin the Software without restriction, including without limitation the rights\nto use, copy, modify, merge, publish, distribute, sublicense, and/or sell\ncopies of the Software, and to permit persons to whom the Software is\nfurnished to do so, subject to the following conditions:\n\nThe above copyright notice and this permission notice shall be included in all\ncopies or substantial portions of the Software.\n\nTHE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR\nIMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,\nFITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE\nAUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER\nLIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,\nOUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE\nSOFTWARE.\n',
    'author': 'walchko',
    'author_email': 'walchko@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pypi.org/project/doop/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)



import numpy as np
from .constants import Earth

def hohmann(r1, r2, mu=Earth.mu):
    #
    # Description: Assuming circular orbits of radius r1 and r2 around the
    # central body described by mu, a hohmann transfer from r1 to r2 is
    # calculated. The total deltaV and time of flight are output.
    #
    # Input:  r1  = radius of orbit 1
    #         r2  = radius of orbit 2
    #         mu  = gravitational parameter of central body
    #
    # Output: dV  = total deltaV of hohmann transfer from r1 to r2 [km/sec]
    #         TOF = total time of flight from orbit r1 to orbit r2 [sec]
    #
    # Calculate the velocities at each circular orbit:
    v1 = np.sqrt(mu/r1)
    v2 = np.sqrt(mu/r2)

    # Calculated the semi-major axis of the transfer orbit:
    a = (r1+r2)/2

    # Calculate the velocities of the transfer orbit:
    v1t = np.sqrt(2*mu/r1 - mu/a)
    v2t = np.sqrt(2*mu/r2 - mu/a)

    # Calculate the deltaV at each r1 and r2:
    dv1 = abs(v1t - v1)
    dv2 = abs(v2t - v2)

    # Calculate the total deltaV and TOF:
    dV = dv1 + dv2
    TOF = np.pi*np.sqrt(a**3/mu)

    return (dV, TOF,)



import numpy as np
from numpy import pi,sqrt
from .constants import Earth

# Add some SMAD calculations here

def period(a):
    # a [km] -> period [sec]
    return 2*pi*sqrt(a**3/Earth.mu)

def dnode(period):
    """
    Orbit nodal precession on each rev. Orbits below GEO move (drift) Westerly
    while orbits above GEO move Easterly. Orbits at GEO are stationary, 0 deg drift.
    Use siderial day rotation for better accuracy.

    Arg:
    period [sec]
    Return:
    node precession (dn) [deg]
    """
    return 360 - 360/(23*3600+56*60+4)*period
